Attribute VB_Name = "modMain"
Option Explicit
Public Declare Function SetPixel Lib "gdi32" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, ByVal crColor As Long) As Long

Public Declare Function InitCommonControlsEx Lib "comctl32.dll" (iccex As tagInitCommonControlsEx) As Boolean
Public Const ICC_USEREX_CLASSES = &H200

Public Type tagInitCommonControlsEx
  lngSize As Long
  lngICC As Long
End Type

Public Const EditorCaption As String = "SML2 MapEditor 1.1 - "

Public Type tMap
  Offset As Long
  GFXOffset As Long
  GFXEndOffset As Long
  IsSubmap As Boolean
End Type

Public TilesetBuffer As cDIBSection, MapBuffer As cDIBSection, ROMFileName As String
Public Maps() As tMap, ROM() As Byte, MapArray() As Byte, Dirty As Boolean
Public TileSelectedX As Integer, TileSelectedY As Integer, StdSelect As Boolean
Public Selecting As Boolean, SelStartX As Integer, SelStartY As Integer
Public SelectWidth As Integer, SelectHeight As Integer, LoadedMapNum As Integer
Public MapSelecting As Boolean, MapSelStartX As Integer, MapSelStartY As Integer
Public MapSelectWidth As Integer, MapSelectHeight As Integer, MapSelected As Boolean
Public MapSelectTiles() As Byte, NoSelection As Boolean

Sub Main()
  Dim iccex As tagInitCommonControlsEx
  
  On Error Resume Next
  
  With iccex
    .lngSize = LenB(iccex)
    .lngICC = ICC_USEREX_CLASSES
  End With
  InitCommonControlsEx iccex
  
  On Error GoTo 0
  Load frmMain
  LoadMapNames
  If Not PickROM Then End
End Sub

Public Function PickROM() As Boolean
  Dim ComDlg As cCommonDialog, ROMName As String
  Dim ROMTitle As String, ROMBackup() As Byte
  
  If Not CheckSave Then Exit Function
  Set ComDlg = New cCommonDialog
  If Not ComDlg.VBGetOpenFileName(ROMName, ROMTitle, , , , , "GameBoy ROMs (*.gb)|*.gb|All Files (*.*)|*.*", , , "Select a Super Mario Land 2 ROM...") Then PickROM = False: Exit Function
  ROMBackup = ROM
  LoadROM ROMName
  If Not IsSML2 Then
    If MsgBox("This may not be a Super Mario Land 2 ROM." + vbLf + "Are you sure you want to load it?" + vbLf + "The editor may crash, corrupt the ROM or behave unexpectedly.", vbYesNo Or vbExclamation, "SML2 MapEditor") = vbNo Then
      ROM = ROMBackup
      Exit Function
    End If
  End If
  LoadMap 0
  frmMain.Caption = EditorCaption + ROMTitle
  frmMain.Show
  PickROM = True
End Function

Public Sub LoadROM(ROMName As String)
  Dim ROMLen As Long, FHandle As Integer
  
  If Dir(ROMName) = "" Then MsgBox "Non-existent ROM file!", vbCritical Or vbOKOnly: Exit Sub
  
  FHandle = FreeFile
  Open ROMName For Binary As #FHandle
    ROMLen = LOF(FHandle)
    ReDim ROM(ROMLen - 1)
    Get #FHandle, , ROM
  Close #FHandle
  ROMFileName = ROMName
End Sub

Public Sub LoadGFX(StartOffset As Long, EndOffset As Long)
  Dim FilePos As Long, DrawX As Long, DrawY As Long
  Dim GBPal(0 To 3) As Long, TileX As Integer, TileY As Integer
  
  GBPal(3) = RGB(0, 0, 0)
  GBPal(2) = RGB(120, 120, 120)
  GBPal(1) = RGB(176, 176, 176)
  GBPal(0) = RGB(255, 255, 255)
  
  Set TilesetBuffer = New cDIBSection
  TilesetBuffer.Create 128, 128
  FilePos = StartOffset
  Do Until FilePos > EndOffset
    For TileX = 7 To 0 Step -1
      SetPixel TilesetBuffer.hdc, DrawX + 7 - TileX, DrawY + TileY, GBPal(IIf(BitIsSet(ROM(FilePos), TileX) = True, 1, 0) + IIf(BitIsSet(ROM(FilePos + 1), TileX) = True, 2, 0))
    Next TileX
    FilePos = FilePos + 2
    TileY = TileY + 1
    If TileY >= 8 Then TileY = 0: DrawX = DrawX + 8
    If DrawX >= 128 Then DrawX = 0: DrawY = DrawY + 8
  Loop
End Sub

Public Sub LoadMap(MapNum As Integer)
  If Not CheckSave Then Exit Sub
  LoadedMapNum = MapNum
  LoadGFX Maps(MapNum).GFXOffset, Maps(MapNum).GFXEndOffset
  LoadMapByOffset Maps(MapNum).Offset, Maps(MapNum).IsSubmap
  RenderMap
  frmMain.cboMap.ListIndex = MapNum
  If frmMain.Visible Then
    frmMain.pbxTileset.Refresh
  End If
  StdSelect = False
  Selecting = False
  MapSelected = False
  MapSelecting = False
  NoSelection = True
  Dirty = False
  frmMain.lblSelection.Caption = "Selection: none"
  frmMain.fraMap.Caption = "Map"
  frmMain.pbxMap.Refresh
  frmMain.pbxTileset.Refresh
End Sub

Public Sub LoadMapByOffset(MapOffset As Long, IsSubmap As Boolean)
  Dim FilePos As Long, ListX As Long, ListY As Long
  Dim TileX As Long, TileY As Long, FileEnd As Long
  
  ReDim MapArray(31, IIf(IsSubmap = True, 19, 31), 1)
  frmMain.pbxMap.Height = IIf(IsSubmap = True, 160, 256) * Screen.TwipsPerPixelY
  Set MapBuffer = New cDIBSection
  MapBuffer.Create 256, IIf(IsSubmap = True, 160, 256)
  FilePos = MapOffset
  FileEnd = MapOffset + IIf(IsSubmap = True, &H27F, &H3FF)
  Do Until FilePos > FileEnd
    TileX = ROM(FilePos) Mod 16
    TileY = ROM(FilePos) \ 16
    TileY = TileY - 8
    If TileY < 0 Then TileY = TileY + 16
    MapArray(ListX, ListY, 0) = TileX
    MapArray(ListX, ListY, 1) = TileY
    FilePos = FilePos + 1
    ListX = ListX + 1
    If ListX >= 32 Then ListX = 0: ListY = ListY + 1
  Loop
End Sub

Public Sub RenderMap()
  Dim TileX As Long, TileY As Long
  Dim ListX As Long, ListY As Long
  Dim DrawX As Long, DrawY As Long
  
  For ListY = 0 To UBound(MapArray, 2)
    For ListX = 0 To 31
      TileX = MapArray(ListX, ListY, 0) * 8
      TileY = MapArray(ListX, ListY, 1) * 8
      TilesetBuffer.PaintPicture MapBuffer.hdc, DrawX, DrawY, 8, 8, TileX, TileY, vbSrcCopy
      DrawX = DrawX + 8
      If DrawX >= 256 Then DrawX = 0: DrawY = DrawY + 8
    Next ListX
  Next ListY
  If frmMain.Visible Then frmMain.pbxMap.Refresh
End Sub

Public Sub PlaceTile(ByVal DestX As Integer, ByVal DestY As Integer, ByVal TileX As Integer, ByVal TileY As Integer)
  If DestX < 0 Or DestX > UBound(MapArray, 1) Or DestY < 0 Or DestY > UBound(MapArray, 2) Then Exit Sub
  MapArray(DestX, DestY, 0) = TileX
  MapArray(DestX, DestY, 1) = TileY
  TilesetBuffer.PaintPicture MapBuffer.hdc, DestX * 8, DestY * 8, 8, 8, TileX * 8, TileY * 8, vbSrcCopy
End Sub

Public Sub LoadMapNames()
  Dim FHandle As Integer, MapNames() As Byte
  Dim Temp As String, MapCount As Integer
  
  If Dir("C:\SML2MAPEDMAPNAMES.txt") <> "" Then Kill "C:\SML2MAPEDMAPNAMES.txt"
  FHandle = FreeFile
  Open "C:\SML2MAPEDMAPNAMES.txt" For Binary As #FHandle
    MapNames = LoadResData("MAP_NAMES", 10)
    Put #FHandle, , MapNames
  Close #FHandle
  
  Open "C:\SML2MAPEDMAPNAMES.txt" For Input As #FHandle
    Do Until EOF(FHandle)
      ReDim Preserve Maps(MapCount)
      With Maps(MapCount)
        Line Input #FHandle, Temp: If Temp = "NEWSUBMAP" Then .IsSubmap = True
        Line Input #FHandle, Temp: frmMain.cboMap.AddItem Temp
        Line Input #FHandle, Temp: .Offset = Val("&H" + Temp)
        Line Input #FHandle, Temp: .GFXOffset = Val("&H" + Temp)
        Line Input #FHandle, Temp: .GFXEndOffset = Val("&H" + Temp)
      End With
      MapCount = MapCount + 1
    Loop
  Close #FHandle
  
  Kill "C:\SML2MAPEDMAPNAMES.txt"
End Sub

Public Sub SaveMap()
  Dim SaveX As Integer, SaveY As Integer, SByte As Byte
  Dim FilePos As Long, FHandle As Integer
  
  Dirty = False
  FilePos = Maps(LoadedMapNum).Offset
  frmMain.fraMap.Caption = "Map"
  For SaveY = 0 To UBound(MapArray, 2)
    For SaveX = 0 To 31
      SByte = MapArray(SaveX, SaveY, 1)
      SByte = SByte + 8
      If SByte > 15 Then SByte = SByte - 16
      SByte = SByte * 16
      SByte = SByte + MapArray(SaveX, SaveY, 0)
      ROM(FilePos) = SByte
      FilePos = FilePos + 1
    Next SaveX
  Next SaveY
  
  FHandle = FreeFile
  Open ROMFileName For Binary As #FHandle
    Put #FHandle, , ROM
  Close #FHandle
End Sub

Public Function IsSML2() As Boolean
  Dim i As Integer, CheckString As String
  
  If UBound(ROM) <> 524287 Then IsSML2 = False: Exit Function
  For i = &H134 To &H13D
    CheckString = CheckString + Chr(ROM(i))
  Next i
  If CheckString = "MARIOLAND2" Then IsSML2 = True Else IsSML2 = False
End Function

Public Function CheckSave() As Boolean
  If Not Dirty Then CheckSave = True: Exit Function
  
  Select Case MsgBox("The map hasn't been saved. Save it?", vbYesNoCancel Or vbQuestion, "SML2 MapEditor")
    Case vbYes
      SaveMap
      CheckSave = True
    Case vbNo
      CheckSave = True
  End Select
End Function

Public Function BitIsSet(ByVal Number As Long, ByVal Bit As Long) As Boolean
  BitIsSet = False
  If Bit = 31 Then
    If Number And &H80000000 Then BitIsSet = True
  Else
    If Number And (2 ^ Bit) Then BitIsSet = True
  End If
End Function
